import { LRLanguage, LanguageSupport } from '@codemirror/language';
import { styleTags, tags } from '@lezer/highlight';
import { LRParser } from '@lezer/lr';

function generateCall(val) {
    return `yield karol.${val}();`;
}
function generateCallWithParam(val, param) {
    return `yield karol.${val}("${param}");`;
}
function generateBoolean(val) {
    return `yield ${val === "wahr"};`;
}
function generateCustomCall(val) {
    return `for(let n of ${val}()){yield n;};`;
}
function generateCallInConditionWithParam(val, param) {
    return `(function*(){yield karol.${val}("${param}")})()`;
}
function generateCallInCondition(val) {
    return `(function*(){yield karol.${val}()})()`;
}
function generateCustomCallInCondition(val) {
    return `${val}()`;
}
function generateInfiniteLoop(body) {
    return `while(true){${body.join("")}};`;
}
function generateWhileLoop(body, cond, inv) {
    return `while(true){let n;for(n of ${cond}){yield n;}if(${inv ? "" : "!"}n){break;}${body.join("")}};`;
}
function generateTimesLoop(body, times) {
    return `for(let i=0;i<${times};i++){${body.join("")}};`;
}
function generateWhileEndLoop(body, cond, inv, bis) {
    return `do{${body.join("")}let n;for(n of ${cond}){yield n;}if(${inv !== bis ? "" : "!"}n){break;}}while(true);`;
}
function generateIfElse(body, elseBody, cond, inv) {
    return `{let n;for(n of ${cond}){yield n;};if(${inv ? "!" : ""}n){${body.join("")}}else{${elseBody.join("")}}};`;
}
function generateIf(body, cond, inv) {
    return `{let n;for(n of ${cond}){yield n;};if(${inv ? "!" : ""}n){${body.join("")}}};`;
}
function generateSub(body, subName) {
    return {
        result: `function* ${subName}(){${body.join("")}};`,
        identifier: subName,
    };
}

function noParamAllowed(pos) {
    return {
        msg: "subroutine/condition calls must not contain parameters",
        pos: pos,
    };
}
function predefinedSubRedefinition(pos) {
    return {
        msg: "redefinition of predefined subroutine",
        pos: pos,
    };
}
function predefinedCondRedefinition(pos) {
    return {
        msg: "redefinition of predefined condition",
        pos: pos,
    };
}
function nestedSubDefintion(pos) {
    return {
        msg: "subroutine must not be declared inside another subroutine/condition declaration",
        pos: pos,
    };
}
function nestedCondDefintion(pos) {
    return {
        msg: "condition must not be declared inside another subroutine/condition declaration",
        pos: pos,
    };
}
function illegalRedefinition(pos) {
    return {
        msg: "illegal redefintion",
        pos: pos,
    };
}

const GermanLanguageSupport = {
    keywords: {
        beginSubroutine: /^(Anweisung|Methode)$/i,
        endSubroutine: /^((ende|\*)(Anweisung|Methode))$/i,
        beginCondition: /^(Bedingung)$/i,
        endCondition: /^((ende|\*)Bedingung)$/i,
        beginIf: /^(wenn)$/i,
        then: /^(dann)$/i,
        beginElse: /^(sonst)$/i,
        endIf: /^((ende|\*)wenn)$/i,
        beginWhile: /^(wiederhole)$/i,
        endWhile: /^((ende|\*)wiederhole)$/i,
        forever: /^(immer)$/i,
        times: /^(mal)$/i,
        doWhile: /^(solange)$/i,
        until: /^(bis)$/i,
        not: /^(nicht)$/i
    },
    colours: {
        red: "rot",
        green: "grün",
        blue: "blau",
        yellow: "gelb",
        black: "schwarz"
    }
};

// This file was generated by lezer-generator. You probably shouldn't edit it.
const beginSubroutine = 1, endSubroutine = 2, beginCondition = 3, endCondition = 4, beginIf = 5, then = 6, beginElse = 7, endIf = 8, beginWhile = 9, endWhile = 10, forever = 11, times = 12, doWhile = 13, until = 14, not = 15, Colour = 16, Program = 17, Identifier = 18, IdentifierWithParam = 20, Number$1 = 22, Subroutine = 23, Condition = 24, If = 25, While = 26, WhileEnd = 27;

var tokens = /*#__PURE__*/Object.freeze({
    __proto__: null,
    beginSubroutine: beginSubroutine,
    endSubroutine: endSubroutine,
    beginCondition: beginCondition,
    endCondition: endCondition,
    beginIf: beginIf,
    then: then,
    beginElse: beginElse,
    endIf: endIf,
    beginWhile: beginWhile,
    endWhile: endWhile,
    forever: forever,
    times: times,
    doWhile: doWhile,
    until: until,
    not: not,
    Colour: Colour,
    Program: Program,
    Identifier: Identifier,
    IdentifierWithParam: IdentifierWithParam,
    Number: Number$1,
    Subroutine: Subroutine,
    Condition: Condition,
    If: If,
    While: While,
    WhileEnd: WhileEnd
});

const tokenMap = new Map(Object.entries(tokens));
function generateTokenizer(langSupport) {
    return (str) => {
        for (const [key, value] of Object.entries(GermanLanguageSupport.keywords))
            if (value.test(str))
                if (tokenMap.has(key))
                    return tokenMap.get(key);
        for (const [key, value] of Object.entries(GermanLanguageSupport.colours))
            if (value === str)
                return Colour;
        return -1;
    };
}
const germanTokenizer = generateTokenizer();

// This file was generated by lezer-generator. You probably shouldn't edit it.
const parser = LRParser.deserialize({
  version: 14,
  states: "*OQYQPOOOkQPO'#C}O!cQPO'#CsO!hQPO'#CtO!mQPO'#CuO!uQPO'#CvOOQO'#C}'#C}OOQO'#Cx'#CxQYQPOOO#VQPO,59[O#_QPO,59_O#fQPO,59`O#mQPO,59aO#uQPO,59aO#zQPO,59aO$PQPO,59bO$WQPO,59bO$]QPO,59bO$eQPO,59cO$mQPO,59cOOQO-E6v-E6vO$tQPO1G.vOOQO1G.y1G.yO$yQPO1G.yOOQO1G.z1G.zO%QQPO1G.zO%XQPO1G.{O%cQPO1G.{O%kQPO1G.{OOQO1G.|1G.|O%pQPO1G.|O%pQPO1G.|O%wQPO1G.|O&RQPO1G.|O&WQPO1G.}O&`QPO1G.}OOQO7+$b7+$bOOQO7+$e7+$eOOQO7+$f7+$fOOQO7+$g7+$gO&hQPO7+$gO&oQPO7+$gO&oQPO7+$gOOQO7+$h7+$hO&yQPO7+$hO'QQPO7+$hO&yQPO7+$hO'[QPO7+$iOOQO7+$i7+$iO(SQPO7+$iO(XQPO7+$iOOQO<<HR<<HRO(aQPO<<HRO(aQPO<<HRO(hQPO<<HROOQO<<HS<<HSO(rQPO<<HSO(yQPO<<HTOOQO<<HT<<HTO)qQPO<<HTOOQOAN=mAN=mO)vQPOAN=mO)vQPOAN=mOOQOAN=nAN=nO)}QPOAN=oOOQOAN=oAN=oOOQOG23XG23XO*uQPOG23XOOQOLD(sLD(s",
  stateData: "*|~OoOSpOS~OPQORROTSOXTObPO~OeXOPqXRqXTqXXqXbqXmqXYqXQqXSqXVqXWqX~ObYO~ObZO~O_^Ob[O~OYbOZ_O]aOf`O~PYO`eOfeO~OQfO~PYOShO~PYOUjOeXO~OUjO~ObkO~OYmO~PYO[oO~O_qObpO~O]rO^rO~OYsO~PYOctO~OQuO~PYOSvO~PYOVxOWwO~PYOUzOeXO~OUzO~OY{O~PYOY{OeXO~PYOb}O~O_!ROb!PO~O]!SO^!SO~OW!TO~PYOV!VOW!TO~PYOY!XO~PYOY!XOeXO~PYOeXOPkqRkqTkqXkqbkqmkqYkqQkqSkqVkqWkq~Ob!ZO~O_!]Ob!ZO~OW!^O~PYOV!`OW!^O~PYOY!aO~PYOeXOPkyRkyTkyXkybkymkyYkyQkySkyVkyWky~Ob!bO~OW!dO~PYOeXOPk!RRk!RTk!RXk!Rbk!Rmk!RYk!RQk!RSk!RVk!RWk!R~OW!fO~PYO",
  goto: "%XrPPPPPPPPPPPPPPPPPPPPsPP!w!w!w!w!w#ePPPP$k!UUOTWYZ_cgijnopxyz|}!O!U!V!W!Y!_!`!eQ]SQl^QoaQ!OqQ!QrS![!R!SR!c!]!VUOTWYZ_cgijnopxyz|}!O!U!V!W!Y!_!`!eQWOQcThdWcginy|!U!W!Y!_!eQgYQiZQn_QyjS|opQ!UxQ!WzS!Y}!OQ!_!VR!e!`!VVOTWYZ_cgijnopxyz|}!O!U!V!W!Y!_!`!e",
  nodeNames: "⚠ Keyword Keyword Keyword Keyword Keyword Keyword Keyword Keyword Keyword Keyword Keyword Keyword Keyword Keyword Keyword Colour Program Identifier ) IdentifierWithParam ( Number Subroutine Condition If While WhileEnd",
  maxTerm: 33,
  skippedNodes: [0],
  repeatNodeCount: 1,
  tokenData: "$n~R`XY!TYZ!T]^!Tpq!Txy!iyz!nz{!s!Q![#b!]!^!T!c!}!s#R#S!s#T#o!s#o#p#j%|%}!s&`&a!s&f&g!s~!YTo~XY!TYZ!T]^!Tpq!T!]!^!T~!nOe~~!sOc~~!xWb~}!O!s!Q![!s!c!}!s#R#S!s#T#o!s%|%}!s&`&a!s&f&g!s~#gPf~!Q![#b~#m^XY#jYZ#j]^#jpq#j}!O#j!Q![#j!]!^#j!c!}#j#R#S#j#T#o#j#q#r$i%|%}#j&`&a#j&f&g#j~$nOp~",
  tokenizers: [0],
  topRules: {"Program":[0,17]},
  specialized: [{term: 18, get: (value, stack) => (germanTokenizer(value, stack) << 1)}],
  tokenPrec: 0
});

const karolLanguage = LRLanguage.define({
    parser: parser.configure({
        props: [
            styleTags({
                Identifier: tags.variableName,
                Colour: tags.literal,
                Number: tags.number,
                comment: tags.blockComment,
                Keyword: tags.keyword,
                "( )": tags.paren,
            }),
        ],
    }),
    languageData: {
        commentTokens: { block: { open: "{", close: "}" } },
    },
});
function karol() {
    return new LanguageSupport(karolLanguage);
}
function parse(str) {
    return parser.parse(str);
}

const callIdentifiers = [
    "schritt",
    "linksdrehen",
    "rechtsdrehen",
    "hinlegen",
    "aufheben",
    "markesetzen",
    "markelöschen",
    "warten",
    "ton",
    "beenden",
];
const conditionIdentifiers = [
    "istwand",
    "nichtistwand",
    "istziegel",
    "nichtistziegel",
    "istmarke",
    "nichtistmarke",
    "istnorden",
    "istosten",
    "istsüden",
    "istwesten",
    "istvoll",
    "nichtistvoll",
    "istleer",
    "nichtistleer",
    "hatziegel",
];

const callIdentifiersSet = new Set(callIdentifiers);
const conditionIdentifiersSet = new Set(conditionIdentifiers);
function getVal(str, cursor) {
    return str.substring(cursor.from, cursor.to);
}
function semanticAnalysis(str, cursor, conditions, subroutines) {
    do {
        let val = getVal(str, cursor);
        let pos = { from: cursor.from, to: cursor.to };
        let lit = cursor.name;
        switch (lit) {
            case "Identifier":
                if (!conditions.has(val) &&
                    !subroutines.has(val) &&
                    !conditionIdentifiersSet.has(val) &&
                    !callIdentifiersSet.has(val) &&
                    val !== "wahr" &&
                    val !== "falsch") {
                    throw {
                        msg: "unknown subroutine/condition",
                        pos: { from: cursor.from, to: cursor.to },
                    };
                }
                break;
            case "IdentifierWithParam":
                break;
            case "If":
                cursor.firstChild();
                cursor.nextSibling();
                val = getVal(str, cursor);
                if (val === "nicht")
                    cursor.nextSibling();
                if (cursor.name === "IdentifierWithParam") {
                    cursor.firstChild();
                    val = getVal(str, cursor);
                    cursor.parent();
                }
                else {
                    val = getVal(str, cursor);
                }
                if (!conditions.has(val) && !conditionIdentifiersSet.has(val)) {
                    throw {
                        msg: "identifier must be a condition",
                        pos: { from: cursor.from, to: cursor.to },
                    };
                }
                cursor.parent();
                break;
            case "While":
                cursor.firstChild(); // wiederhole
                cursor.nextSibling(); // "immer" | "solange" | Number
                val = getVal(str, cursor);
                if (val === "immer" || cursor.name === "Number") {
                    // skip times and forever loops
                    cursor.parent();
                    break;
                }
                cursor.nextSibling();
                if (getVal(str, cursor) === "nicht")
                    cursor.nextSibling();
                if (cursor.name === "IdentifierWithParam") {
                    cursor.firstChild();
                    val = getVal(str, cursor);
                    cursor.parent();
                }
                else {
                    val = getVal(str, cursor);
                }
                if (!conditions.has(val) && !conditionIdentifiersSet.has(val)) {
                    throw {
                        msg: "identifier must be a condition",
                        pos: { from: cursor.from, to: cursor.to },
                    };
                }
                cursor.parent();
                break;
            case "WhileEnd":
                cursor.firstChild(); // "wiederhole"
                while (cursor.nextSibling()) {
                    let val = getVal(str, cursor);
                    if (val === "endewiederhole" || val === "*wiederhole")
                        break;
                }
                cursor.nextSibling(); // "solange" | "bis"
                cursor.nextSibling();
                if (getVal(str, cursor) === "nicht") {
                    cursor.nextSibling();
                }
                if (cursor.name === "IdentifierWithParam") {
                    cursor.firstChild();
                    val = getVal(str, cursor);
                    cursor.parent();
                }
                else {
                    val = getVal(str, cursor);
                }
                if (!conditions.has(val) && !conditionIdentifiersSet.has(val)) {
                    throw {
                        msg: "identifier must be a condition",
                        pos: { from: cursor.from, to: cursor.to },
                    };
                }
                cursor.parent();
                break;
            case "Subroutine":
                break;
            case "Condition":
                break;
            case "Keyword":
                break;
            case "Number":
                break;
            case "Colour":
                break;
            case "(":
                break;
            case ")":
                break;
            default:
                // faulty node detected -> parser error
                throw { msg: "parse error", pos: pos };
        }
    } while (cursor.next());
}

function isPredefinedIdentifier(val) {
    return callIdentifiersSet.has(val) || conditionIdentifiersSet.has(val);
}
function isBool(val) {
    return val === "wahr" || val === "falsch";
}
function isEnde(token, val) {
    return val === `ende${token}` || val === `*${token}`;
}
function isNotRedefined(subroutines, conditions, defRes) {
    return (!subroutines.has(defRes.identifier) && !conditions.has(defRes.identifier));
}
function compileIdentifier(str, cursor) {
    let val = getVal(str, cursor);
    let pos = { from: cursor.from, to: cursor.to };
    if (cursor.name === "IdentifierWithParam") {
        cursor.firstChild(); // Identifier
        val = getVal(str, cursor);
        cursor.nextSibling(); // "("
        cursor.nextSibling(); // (Colour | Number)
        let param = getVal(str, cursor);
        if (isPredefinedIdentifier(val)) {
            cursor.parent(); // IdentifierWithParam
            return generateCallWithParam(val, param);
        }
        else {
            cursor.parent(); // IdentifierWithParam
            throw noParamAllowed(pos);
        }
    }
    else {
        if (isPredefinedIdentifier(val)) {
            return generateCall(val);
        }
        if (isBool(val)) {
            return generateBoolean(val);
        }
        return generateCustomCall(val);
    }
}
function compileConditionIdentifier(str, cursor) {
    let val = getVal(str, cursor);
    let pos = { from: cursor.from, to: cursor.to };
    if (cursor.name === "IdentifierWithParam") {
        cursor.firstChild(); // Identifier
        val = getVal(str, cursor);
        cursor.nextSibling(); // "("
        cursor.nextSibling(); // (Colour | Number)
        let param = getVal(str, cursor);
        if (isPredefinedIdentifier(val)) {
            cursor.parent(); // IdentifierWithParam
            return generateCallInConditionWithParam(val, param);
        }
        else {
            cursor.parent(); // IdentifierWithParam
            throw noParamAllowed(pos);
        }
    }
    else {
        if (isPredefinedIdentifier(val)) {
            return generateCallInCondition(val);
        }
        return generateCustomCallInCondition(val);
    }
}
function compileWhile(str, cursor) {
    cursor.firstChild(); // beginWhile
    cursor.nextSibling(); // (forever | Number times | doWhile not? (Identifier | IdentifierWithParam))
    let condType = getVal(str, cursor);
    let inv = false;
    let times = 0;
    let cond = "";
    if (condType === "solange") {
        cursor.nextSibling(); // not?
        if (getVal(str, cursor) === "nicht") {
            cursor.nextSibling();
            inv = true;
        }
        cond = compileConditionIdentifier(str, cursor);
    }
    else if (cursor.name === "Number") {
        times = Number(condType);
        cursor.nextSibling(); // times
    }
    let body = [];
    while (cursor.nextSibling()) {
        let val = getVal(str, cursor);
        if (isEnde("wiederhole", val))
            break;
        let res = compileInner(str, cursor);
        body.push(res);
    }
    cursor.parent(); // While
    if (condType === "immer") {
        return generateInfiniteLoop(body);
    }
    else if (condType === "solange") {
        return generateWhileLoop(body, cond, inv);
    }
    else {
        return generateTimesLoop(body, times);
    }
}
function compileWhileEnd(str, cursor) {
    cursor.firstChild(); // beginWhile
    let body = [];
    while (cursor.nextSibling()) {
        let val = getVal(str, cursor);
        if (isEnde("wiederhole", val))
            break;
        let res = compileInner(str, cursor);
        body.push(res);
    }
    cursor.nextSibling(); // "solange" | "bis"
    let condType = getVal(str, cursor);
    let inv = false;
    let bis = false;
    let cond;
    if (condType === "bis") {
        bis = true;
    }
    cursor.nextSibling(); // not?
    if (getVal(str, cursor) === "nicht") {
        cursor.nextSibling();
        inv = true;
    }
    cond = compileConditionIdentifier(str, cursor);
    cursor.parent(); // WhileEnd
    return generateWhileEndLoop(body, cond, inv, bis);
}
function compileIf(str, cursor) {
    cursor.firstChild(); // "beginIf"
    cursor.nextSibling();
    let val = getVal(str, cursor);
    let inv = false;
    if (val === "nicht") {
        cursor.nextSibling();
        inv = true;
    }
    let cond = compileConditionIdentifier(str, cursor);
    cursor.nextSibling(); // "dann"
    let isElse = false;
    const body = [];
    const elseBody = [];
    while (cursor.nextSibling()) {
        val = getVal(str, cursor);
        if (isEnde("wenn", val))
            break;
        if (val === "sonst") {
            isElse = true;
            cursor.nextSibling();
        }
        if (isElse) {
            let res = compileInner(str, cursor);
            elseBody.push(res);
        }
        else {
            let res = compileInner(str, cursor);
            body.push(res);
        }
    }
    cursor.parent();
    if (isElse)
        return generateIfElse(body, elseBody, cond, inv);
    return generateIf(body, cond, inv);
}
function compileDefinition(str, cursor) {
    let pos = { from: cursor.from, to: cursor.to };
    let isSub = cursor.name === "Subroutine";
    cursor.firstChild();
    cursor.nextSibling();
    let subName = getVal(str, cursor);
    if (isPredefinedIdentifier(subName))
        if (isSub)
            throw predefinedSubRedefinition(pos);
        else
            throw predefinedCondRedefinition(pos);
    let val;
    const body = [];
    while (cursor.nextSibling()) {
        val = getVal(str, cursor);
        if (isEnde(isSub ? "anweisung" : "bedingung", val))
            break;
        let res = compileInner(str, cursor);
        body.push(res);
    }
    cursor.parent();
    return generateSub(body, subName);
}
function compileInner(str, cursor) {
    getVal(str, cursor);
    let pos = { from: cursor.from, to: cursor.to };
    switch (cursor.name) {
        case "Identifier":
            return compileIdentifier(str, cursor);
        case "IdentifierWithParam":
            return compileIdentifier(str, cursor);
        case "If":
            return compileIf(str, cursor);
        case "While":
            return compileWhile(str, cursor);
        case "WhileEnd":
            return compileWhileEnd(str, cursor);
        case "Subroutine":
            throw nestedSubDefintion(pos);
        case "Condition":
            throw nestedCondDefintion(pos);
        default:
            // faulty node detected -> parser error
            throw { msg: "parse error", pos: pos };
    }
}
function compile(str) {
    str = str.toLowerCase();
    let cursor = parse(str).cursor();
    const program = [];
    const conditions = new Set();
    const subroutines = new Set();
    cursor.firstChild();
    do {
        getVal(str, cursor);
        let pos = { from: cursor.from, to: cursor.to };
        let res;
        let defRes;
        if (cursor.name === "Subroutine" || cursor.name === "Condition") {
            defRes = compileDefinition(str, cursor);
            if (isNotRedefined(subroutines, conditions, defRes))
                if (cursor.name === "Subroutine")
                    subroutines.add(defRes.identifier);
                else
                    conditions.add(defRes.identifier);
            else
                throw illegalRedefinition(pos);
            program.push(defRes.result);
        }
        else {
            res = compileInner(str, cursor);
            program.push(res);
        }
    } while (cursor.nextSibling());
    cursor.parent();
    cursor.firstChild();
    semanticAnalysis(str, cursor, conditions, subroutines);
    let GeneratorFunction = Object.getPrototypeOf(function* () { }).constructor;
    return new GeneratorFunction("karol", `${program.join("")}`);
}

export { callIdentifiers, callIdentifiersSet, compile, conditionIdentifiers, conditionIdentifiersSet, karol, karolLanguage, parse };
